/*
    armlcd.c - LCD display emulation in an X window.
    ARMulator extensions for the ARM7100 family.
    Copyright (C) 1999  Ben Williamson

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <assert.h>
/* chy change 2002-12-3
#include <gtk/gtk.h>
*/

#include "armdefs.h"


#define MAX_DEPTH	4		/* bits per pixel */
#define GREY_LEVELS	16
#define LCD_BASE	0xC0000000

/*chy change 2002-12-3*/
#define NO_LCD

#ifdef NO_LCD
void	lcd_enable(ARMul_State *state, int width, int height, int depth){};
void	lcd_disable(ARMul_State *state) {};
void	lcd_write(ARMul_State *state, ARMword addr, ARMword data){};
void	lcd_cycle(ARMul_State *state){};
void	lcd_dirty(ARMul_State *state, ARMword addr){};
void	lcd_clean(ARMul_State *state){};

#else
static GtkWidget *window;
static GdkGC *gc[GREY_LEVELS];


static int no_X = 0;
static int lcd_width, lcd_height, lcd_depth;
static int lcd_palette[GREY_LEVELS] = {0, 15, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13};

static gint
expose_event (GtkWidget *widget, GdkEventExpose *event)
{
	/* event->area.x, event->area.y, event->area.width, event->area.height */
	return FALSE;
}

static gint
button_press_event (GtkWidget *widget, GdkEventButton *event)
{
	/* event->x, event->y, event->button */
	return TRUE;
}

static gint
motion_notify_event (GtkWidget *widget, GdkEventMotion *event)
{
	/* event->x, event->y */
	return TRUE;
}


extern int global_argc;
extern char **global_argv;


void
lcd_cycle(ARMul_State *state)
{
	gtk_main_iteration_do(FALSE);
}

void
lcd_enable(ARMul_State *state, int width, int height, int depth)
{

	int i;
	static int once = 0;
	GdkColormap *colormap;

	if (getenv("DISPLAY") == NULL) {
		no_X++;
		return;
	}

	if (!once) {
		once++;
		gtk_init(&global_argc, &global_argv);
	}

	lcd_width = width;
	lcd_height = height;
	lcd_depth = depth;
	state->io.lcd_limit = LCD_BASE + (width * height * depth / 8);
	
	window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_widget_set_name(window, "DXTree");
	gtk_widget_set_usize(window, width, height);

	gtk_signal_connect(GTK_OBJECT(window), "expose_event",
						(GtkSignalFunc)expose_event, NULL);
	gtk_signal_connect(GTK_OBJECT(window), "motion_notify_event",
						(GtkSignalFunc)motion_notify_event, NULL);
	gtk_signal_connect(GTK_OBJECT(window), "button_press_event",
						(GtkSignalFunc)button_press_event, NULL);

	gtk_widget_set_events(window, GDK_EXPOSURE_MASK
								 | GDK_LEAVE_NOTIFY_MASK
								 | GDK_BUTTON_PRESS_MASK
								 | GDK_POINTER_MOTION_MASK
								 | GDK_POINTER_MOTION_HINT_MASK);
	gtk_widget_show(window);

	colormap = gdk_window_get_colormap(window->window);
	for (i = 0; i < GREY_LEVELS; i++) {
		GdkColor color;

		/* #7fc5a7 is a nice LCDish colour */

		color.red = (15 - i) * (0x7f00 / GREY_LEVELS);
		color.green = (15 - i) * (0xc500 / GREY_LEVELS);
		color.blue = (15 - i) * (0xa700 / GREY_LEVELS);
		gdk_color_alloc (colormap, &color);
		gc[i] = gdk_gc_new(window->window);
		gdk_gc_set_foreground(gc[i], &color);
	}
}


void
lcd_disable(ARMul_State *state)
{
	int i;
	
	if (no_X)
		return;

	for (i = 0; i < GREY_LEVELS; i++) {
		if (gc[i]) {
			gdk_gc_destroy(gc[i]);
			gc[i] = NULL;
		}
	}
	if (window) {
		gtk_widget_destroy(window);
		window = NULL;
	}
}

void
lcd_write(ARMul_State *state, ARMword addr, ARMword data)
{
	ARMword offset;
	int pixnum, x, y, bit;
	
	if (no_X)
		return;

	offset = (addr & ~3) - LCD_BASE;
	pixnum = offset * 8 / lcd_depth;
	x = pixnum % lcd_width;
	y = pixnum / lcd_width;
	assert(y < lcd_height);
	
	for (bit = 0; bit < 32; bit += lcd_depth) {
		/* XXX need palette code here */
		int pix = (data >> bit) % (1 << lcd_depth);
		gdk_draw_point(window->window, gc[lcd_palette[pix]], x, y);
		x++;
	}
}


#endif

