/*
	skyeye_mach_at91.c - define machine at91 for skyeye
	Copyright (C) 2003 Skyeye Develop Group
	for help please send mail to <skyeye-developer@lists.sf.linuxforum.net> 
	
	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 
*/
/*
 * 3/24/2003 	init this file.
 * 		add machine at91's function.Most taken from original armio.c 
 * 		include: at91_mach_init, at91_io_do_cycle
 * 		at91_io_read_word, at91_io_write_word
 *		walimis <walimi@peoplemail.com.cn> 
 *		
 *3/24/2003	chenyu <chenyu-tmlinux@hpclab.cs.tsinghua.edu.cn> has done the
 *		necessary work to armio.c
 * */
#include "armdefs.h"
#include "at91.h"
#include "skyeye-ne2k.h"
extern int skyeye_net_maxfd;
extern int skyeye_net_on;
//zzc:2005-1-1
#ifdef __CYGWIN__
#include "mpw/sys/time.h"
#endif

//ywc 2004-04-01
extern unsigned long Pen_buffer[8]; // defined in skyeye_lcd.c

#define TC_DIVISOR	(50)	/* Set your BogoMips here :) may move elsewhere*/

#define DEBUG 0
#if DEBUG
#define DBG_PRINT(a...) fprintf(stderr, ##a)
#else
#define DBG_PRINT(a...)
#endif

typedef struct at91_io {
        ARMword         syscon;                 /* System control */
        ARMword         sysflg;                 /* System status flags */
        ARMword         intsr;                  /* Interrupt status reg */
        ARMword         intmr;                  /* Interrupt mask reg */
        ARMword         tcd[2];                 /* Timer/counter data */
        ARMword         tcd_reload[2];          /* Last value written */
        int             tc_prescale;
        ARMword         uartdr;                 /* Receive data register */
	/*zy 2004-4-13 */
        ARMword         lcdcon;                 /* LCD control */
        //ARMword         lcd_limit;              /* 0xc0000000 <= LCD buffer < lcd_limit */
	ARMword		lcd_is_enable;		/*LCD is turn on"*/
	ARMword		lcd_addr_begin;		/*the begin display mem of LCD*/
	ARMword		lcd_addr_end;		/* the end display of LCD*/ 
	
	//ywc,2004-04-01
	ARMword		ts_int;                     /* ywc 2004-04-02 */
	ARMword		ts_buffer[8];
	//ARMword		tsaddr;                	/* ??? touch srceen data buffer start address*/
	ARMword 	ts_is_enable;
	ARMword 	ts_addr_begin;
	ARMword 	ts_addr_end;
	
	/*ksh 2004-3-6, for net supported*/
        ARMword		net_int;			/*netcard interrupt bit*/
	ARMword		net_io_baseaddr; /*netcard iomem start addr*/
	ARMword		net_io_size; /*netcard iomem size */

} at91_io_t;
static at91_io_t at91_io;
#define io at91_io
/* zy 2004-4-13 */
static void at91_lcd_write(ARMul_State *state, ARMword addr, ARMword data)
{
	        lcd_write(state, addr, data);
}

static void at91_lcd_disable(ARMul_State *state)
{
	        lcd_disable(state);
}
static void at91_lcd_enable(ARMul_State *state, int width, int height, int depth)
{
	        
		lcd_enable(state, width, height, depth);
		
}
static void at91_update_lcd(ARMul_State *state)
{
	        at91_lcd_disable(state);
		
	        if (io.syscon & LCDEN) {
	                ARMword lcdcon = io.lcdcon;
	                ARMword vbufsiz = lcdcon & VBUFSIZ;
	                ARMword linelen = (lcdcon & LINELEN) >> LINELEN_SHIFT;
	                int width, height, depth;
	                switch (lcdcon & (GSEN|GSMD)) {
		                case GSEN:
		                        depth = 2;
	        	                break;
		                case GSEN|GSMD:
		                        depth = 4;
		                        break;
			        default:
			                depth = 1;
			                break;
			}
			depth=8;
			width = (linelen + 1) * 16;
			height = (vbufsiz + 1) * 128 / depth / width;
			at91_lcd_enable(state, width, height, depth);
			
		}
}
//
static void at91_update_int(ARMul_State *state)
{
        ARMword requests = io.intsr & io.intmr;
        state->NfiqSig = (requests & 0x0001) ? LOW : HIGH;
        state->NirqSig = (requests & 0xfffe) ? LOW : HIGH;
}
static void at91_io_reset(ARMul_State *state)
{
        io.syscon = 0;
        io.sysflg = URXFE;
        io.intmr = 0;
        io.intsr = UTXINT;      /* always ready to transmit */
        io.tcd[0] = 0xffff;
        io.tcd[1] = 0xffff;
        io.tcd_reload[0] = 0xffff;
        io.tcd_reload[1] = 0xffff;
        io.tc_prescale = TC_DIVISOR;
        io.uartdr = 0;
        io.lcdcon = 0;
	//zy 2004-4-13
	io.lcd_is_enable           =1;
	//io.lcd_is_enable           =0;
	io.lcd_addr_begin  =0xc0000000;
	io.lcd_addr_end    =0xc0000000;
	//io.lcd_addr_begin  =0x02000000;
	//io.lcd_addr_end    =0x02000000;
	
	//ywc 2004-04-01
	//io.tsaddr		=0x0;
	//io.ts_int		=EINT2; /* ywc 2004-04-02 use EINT2 as touch screen interrupt */
	io.ts_int		=AT91_EXT1; /* ywc 2004-04-02 use EINT2 as touch screen interrupt */
	io.ts_is_enable		=1;
	//io.ts_addr_begin  	=0x8000b000;
	//io.ts_addr_end    	=0x8000b01f;
	io.ts_addr_begin  	=0xff00b000;
	io.ts_addr_end    	=0xff00b01f;
	
	 io.net_int = AT91_EXT0;
	 io.net_io_baseaddr = 0xfffa0000;
	 io.net_io_size = 255;
	
        //state->Exception = TRUE;
}
static ARMword rpr;
int rcr;

/*at91 io_do_cycle*/
void at91_io_do_cycle(ARMul_State *state)
{
	int t;

	io.tc_prescale--;
	if (io.tc_prescale < 0) {
       		io.tc_prescale = TC_DIVISOR;

                for (t = 0; t < 2; t++) {
                        if (io.tcd[t] == 0) {

                                if (io.syscon & (t ? TC2M : TC1M)) {

                                        io.tcd[t] = io.tcd_reload[t];
                                } else {
                                        io.tcd[t] = 0xffff;
                                }
                                io.intsr |= (t ? TC2OI : TC1OI);
                                at91_update_int(state);
                        } else {
                                io.tcd[t]--;
                        }
                }
		if (rcr > 0 && !(io.intsr & URXINT)) {
                        fd_set rfds;
                        struct timeval tv;

                        FD_ZERO(&rfds);
                        FD_SET(skyeye_config.uart.fd_in, &rfds);
                        tv.tv_sec = 0;
                        tv.tv_usec = 0;

                        if (select(skyeye_config.uart.fd_in+1, &rfds, NULL, NULL, &tv) == 1) {
                                unsigned char buf[16];
                                int n, i;
                       		n = read(skyeye_config.uart.fd_in, buf, 
								rcr > sizeof(buf) ? sizeof(buf) : rcr);
                                for (i = 0; i < n; i++, rcr--, rpr++)
                                        mem_write_char(state, rpr, buf[i]);
                                if (rcr == 0)
                                        io.sysflg &= ~URXFE;
                                if (n) {
                                        io.intsr |= URXINT;
                                        at91_update_int(state);
                                }
                        }
                }/* if (rcr > 0 && ...*/
		{  
		int ret;
		if(skyeye_net_on && !(io.intsr & io.net_int )){ 
//chy:if no net irq, then test the incoming packet from tap
			fd_set frds;
			struct timeval tv;
            		FD_ZERO(&frds);
            		//FD_SET(tapif_fd, &frds);
            		skyeye_net_fdset(&frds);
			tv.tv_sec = 0;
			tv.tv_usec = 0;
			ret=0;
			if((ret = select(skyeye_net_maxfd+1,&frds,NULL,NULL,&tv))>0) {
			        skyeye_net_input(&frds,state);
				/*
				if(FD_ISSET(tapif_fd,&frds)){
                			tapif_input(state);
                	        }
				*/
			}
           	}

        	}
		
		// ywc 2004-04-01 for touch screen interrupt
		{
			// ywc 2004-04-01
             
		if(!(io.intsr&io.ts_int)){ //if now has no ts interrupt,then query 
	              if(Pen_buffer[6]==1){  //should trigger a interrupt
				*(state->mach_io.ts_buffer+0)=Pen_buffer[0];
				*(state->mach_io.ts_buffer+1)=Pen_buffer[1];
		//		*(state->mach_io.ts_buffer+2)=Pen_buffer[2];
		//		*(state->mach_io.ts_buffer+3)=Pen_buffer[3];
				*(state->mach_io.ts_buffer+4)=Pen_buffer[4];
		//		*(state->mach_io.ts_buffer+5)=Pen_buffer[5];
				*(state->mach_io.ts_buffer+6)=Pen_buffer[6];
		//		*(state->mach_io.ts_buffer+7)=Pen_buffer[7];
	/*				
				printf("Skyeye: skyeye_mach_at91.c: at91_io_do_cycle(): *(state->mach_io.ts_buffer+0)=%d\n",*(state->mach_io.ts_buffer+0));
				printf("Skyeye: skyeye_mach_at91.c: at91_io_do_cycle(): *(state->mach_io.ts_buffer+1)=%d\n",*(state->mach_io.ts_buffer+1));
				printf("Skyeye: skyeye_mach_at91.c: at91_io_do_cycle(): *(state->mach_io.ts_buffer+2)=%d\n",*(state->mach_io.ts_buffer+2));
				printf("Skyeye: skyeye_mach_at91.c: at91_io_do_cycle(): *(state->mach_io.ts_buffer+3)=%d\n\n",*(state->mach_io.ts_buffer+3));
	*/		        
        			//printf("\nbefore set ENIT2 bit,io.intsr=%x",io.intsr);	
				io.intsr|= io.ts_int;//set EINT2 bit to trigger a interrupt,ts driver will clear it
        			//printf("\nafter set ENIT2 bit,io.intsr=%x\n",io.intsr);	
				
			Pen_buffer[6]=0;
        		}
		 }
	     }
		
		/* zy 2004-4-13*/
		if (io.lcd_is_enable)
			lcd_cycle(state);
	//
	}/* if (io.tc_prescale < 0)*/
}

ARMword	fix_int(ARMword val)
{
	ARMword ret = 0;
	if (val & (1 << 2))
		ret |= URXINT;
	if (val & (1 << 5))
		ret |= TC1OI;
	if (val & (1 << 6))
		ret |= TC2OI;
	if (val & (1 << 16))
		ret |= AT91_EXT0;
        //ywc 2004-05-21 
	if (val & (1 << 17))
		ret |= AT91_EXT1;
	return(ret);
}

ARMword	unfix_int(ARMword val)
{
	ARMword ret = 0;
	if (val & URXINT)
		ret |= (1 << 2);
	if (val & TC1OI)
		ret |= (1 << 5);
	if (val & TC2OI)
		ret |= (1 << 6);
	if (val & AT91_EXT0)
		ret |= (1 << 16);
	//ywc 2004-05-21
	if (val & AT91_EXT1)
		ret |= (1 << 17);
	return(ret);
}
/*
 *  * Atmel serial support for uClinux console.
 *   */

static char *uart_reg[] = {
	"CR",
	"MR",
	"IER",
	"IDR",
	"IMR",
	"CSR",
	"RHR",
	"THR",
	"BRGR",
	"RTOR",
	"TTGR",
	"RES1",
	"RPR",
	"RCR",
	"TPR",
	"TCR",
};
ARMword
uart_read(ARMul_State *state, ARMword addr)
{
	ARMword data;

	switch ((addr & 0xfff) >> 2) {
	case 0xd: // RCR
		data = rcr;
		break;
	case 0xf: // TCR
		data = 0;
		break;
	case 0x5: // CSR
		data = (1<<9) | (1<<1) | /* always empty and ready */
				(rcr ? ((1 << 3) | (1 << 0)) : 0);
		break;
	case 0x0: // CR
	case 0x1: // MR
	case 0x2: // IER
	case 0x3: // IDR
	case 0x4: // IMR
	case 0x6: // RHR
	case 0x7: // THR
	case 0x8: // BRGR
	case 0x9: // RTOR
	case 0xa: // TTGR
	case 0xb: // RES1
	case 0xc: // RPR
	case 0xe: // TPR
		DBG_PRINT("uart_read(%s=0x%08x)\n", uart_reg[(addr&0xfff) >> 2], addr);
		break;
	}
	return(data);
}


void
uart_write(ARMul_State *state, ARMword addr, ARMword data)
{
	static ARMword tx_buf = 0;

	DBG_PRINT("uart_write(0x%x, 0x%x)\n", (addr & 0xfff) >> 2, data);
	switch ((addr & 0xfff) >> 2) {
	case 0x0: // CR
	case 0x3: // IDR
	case 0x9: // RTOR
	case 0x2: // IER
		break;
	case 0x1: // MR
	case 0x4: // IMR
	case 0x5: // CSR
	case 0x6: // RHR
		DBG_PRINT("uart_write(%s=0x%x) = 0x%08x\n",
				uart_reg[(addr & 0xfff) >> 2], addr, data);
		break;
	case 0x7: // THR
		{
			char c = data;
		        write(skyeye_config.uart.fd_out, &c, 1);
		}
		//printf("%c", data); fflush(stdout);
		break;
	case 0x8: // BRGR
	case 0xa: // TTGR
	case 0xb: // RES1
		DBG_PRINT("uart_write(%s=0x%x) = 0x%08x\n",
				uart_reg[(addr & 0xfff) >> 2], addr, data);
		break;
	case 0xd: // RCR
		rcr = data;
		if (rcr == 0)
			io.sysflg &= ~URXFE;
		else
			io.sysflg |= URXFE;
		break;
	case 0xc: // RPR
		DBG_PRINT("uart_write(%s=0x%x) = 0x%08x\n",
				uart_reg[(addr & 0xfff) >> 2], addr, data);
		rpr = data;
		break;
	case 0xe: // TPR
		tx_buf = data;
		break;
	case 0xf: // TCR
		for (; tx_buf && data > 0; data--)
		{
			char c = mem_read_char(state, tx_buf++);
		        write(skyeye_config.uart.fd_out, &c, 1);
			//printf("%c", mem_read_char(state, tx_buf++));
			//fflush(stdout);
		}
		tx_buf = 0;
		break;
	}
}

ARMword at91_io_read_byte(ARMul_State *state, ARMword addr)
{
		ARMword data= -1;
		unsigned char offset=0;
        	unsigned char ret;
		if(addr >= io.net_io_baseaddr && addr < (io.net_io_baseaddr + io.net_io_size)){
                	offset = addr - io.net_io_baseaddr;
                	DBG_PRINT("nic_read begin, offset 0x%x\n",offset);
               	//yangye 2003-1-20 
		        //chy 2003-04-26, only support one card, so the nic index must be 0
                	ret = nic_read(0,state,offset>>2);
                	DBG_PRINT("nic_read end, offset 0x%x, ret %x\n",offset>>2,ret);
                	data = ret;
                	return data;
        	}else{
			printf("SKYEYE: at91_io_read_byte error\n");
			SKYEYE_OUTREGS(stderr);
			exit(-1);
		}
}

ARMword at91_io_read_halfword(ARMul_State *state, ARMword addr)
{
	printf("SKYEYE: at91_io_read_halfword error\n");
	SKYEYE_OUTREGS(stderr);
	exit(-1);
}

ARMword at91_io_read_word(ARMul_State *state, ARMword addr)
{
	/*
 * 	 * The Atmel system registers
 * 	 	 */

	ARMword data = -1;
	static ARMword current_ivr = 0; /* mega hack,  2.0 needs this */
	int i;
	ARMword dataimr = 0;
	//ywc 2004-04-01 read the touch srceen data buffer,return to the ts device driver
	ARMword ts_addr;


	if(addr >= io.net_io_baseaddr && addr < (io.net_io_baseaddr + io.net_io_size)){
		printf("SKYEYE: at91_io_read_word error, shoud use read_byte for net\n");
		SKYEYE_OUTREGS(stderr);
		exit(-1);
	}
#if 0
	{
		unsigned char offset=0;
        	unsigned char ret;
		if(addr >= io.net_io_baseaddr && addr < (io.net_io_baseaddr + io.net_io_size)){
                	offset = addr - io.net_io_baseaddr;
                	DBG_PRINT("nic_read begin, offset 0x%x\n",offset);
               	//yangye 2003-1-20 
		        //chy 2003-04-26, only support one card, so the nic index must be 0
                	ret = nic_read(0,state,offset>>2);
                	DBG_PRINT("nic_read end, offset 0x%x, ret %x\n",offset>>2,ret);
                	data = ret;
                	return data;
        	}
	}
#endif
	ts_addr=addr & ~3;   // 1 word==4 byte
	if(ts_addr >= io.ts_addr_begin && ts_addr <= io.ts_addr_end){
		data=io.ts_buffer[(ts_addr-io.ts_addr_begin)/4];
	//	printf("SKYEYE:ep7312_io_read_word:\n");
        //        printf("addr=%p,ts_addr=%p\n",addr,ts_addr);
                //printf("&io.ts_buffer[%d]=%p\n",(ts_addr-io.ts_addr_begin)/4,&io.ts_buffer[(ts_addr-io.ts_addr_begin)/4]);
          //      printf("data=%d\n",data);
		return data;
	}
	switch (addr) {
	case 0xfffff100: /* IVR */
		data = unfix_int(io.intsr);
		DBG_PRINT("IVR irqs=%x ", data);
		dataimr = unfix_int(io.intmr);
		for (i = 0; i < 32; i++)
			if (data & (1 << i))
				break;
		if (i < 32) {
			data = i;
			io.intsr &= ~fix_int(1 << data);
			at91_update_int(state);
		} else
			data = 0;
		current_ivr = data;
		DBG_PRINT("read IVR=%d\n", data);
		break;
	case 0xfffff108: /* interrupt status register */
		data = current_ivr;
		DBG_PRINT("read ISR=%x,%x\n", data, io.intsr);
		break;
	case 0xfffff110: /* IMR */
		data = unfix_int(io.intmr);
		DBG_PRINT("read IMR=%x,%x\n", data,io.intmr);
		break;
	case 0xfffff114: /* CORE interrupt status register */
		data = unfix_int(io.intsr);
		DBG_PRINT("read CISR=%x,%x\n", data,io.intsr);
		break;
	case 0xfffff120: /* IECR */
		data = unfix_int(io.intmr);
		DBG_PRINT("read IECR=%x,%x\n", data,io.intmr);
		break;
	case 0xfffff124: /* IDCR */
		break;
	case 0xfff00000: /* CPU ID */
		data =  0x2078aa0;
		data = 0x14000040;
		break;
	case 0xfffe00c0: /* TIMER 1 BCR */
		DBG_PRINT("T1-BCR io_read_word(0x%08x) = 0x%08x\n", addr, data);
		break;
	case 0xfffe00c4: /* TIMER 1 BMR */
		DBG_PRINT("T1-BMR io_read_word(0x%08x) = 0x%08x\n", addr, data);
		break;
	case 0xfffe0040: /* TIMER 1 CCR */
		DBG_PRINT("T1-CCR io_read_word(0x%08x) = 0x%08x\n", addr, data);
		break;
	case 0xfffe0044: /* TIMER 1 CMR */
		DBG_PRINT("T1-CMR io_read_word(0x%08x) = 0x%08x\n", addr, data);
		break;
	case 0xfffe0068: /* TIMER 1 IDR */
		DBG_PRINT("T1-IDR io_read_word(0x%08x) = 0x%08x\n", addr, data);
		break;
	case 0xfffe0050: /* TIMER 1 CV */
		DBG_PRINT("T1-CV io_read_word(0x%08x) = 0x%08x\n", addr, data);
		data = io.tcd[0];
		break;
	case 0xfffe005c: /* TIMER 1 RC */
		DBG_PRINT("T1-RC io_read_word(0x%08x) = 0x%08x\n", addr, data);
		data = io.tcd[0];
		break;
	case 0xfffe0060: /* TIMER 1 SR */
		DBG_PRINT("T1-SR io_read_word(0x%08x) = 0x%08x\n", addr, data);
		io.intsr &= ~TC1OI;
		break;
	case 0xfffe0064: /* TIMER 1 IER */
		DBG_PRINT("T1-IER io_read_word(0x%08x) = 0x%08x\n", addr, data);
		break;
	/*zy 2004-4-13 */
	case 0xffc00100: /*SYSCON     	*/ /* you may like to change it ,but it's only for simulation*/
		data = io.syscon;
		break;
	case 0xffc002c0: /*LCDCON	*/ /* because AT91 hasn't LCD support ,so it's just for simulation*/
					   /* you can change this address to what may be suitable for your board */
		data = io.lcdcon;
		break;
	//
	case 0xffc00240: //INTSR
		data = io.intsr;
		break;	
	default:
		if ((addr & 0xffffff00) == 0xfffff000)
			break;
		if ((addr & 0xfffff000) == 0xfffd0000) {
			data = uart_read(state, addr);
			break;
		}
		if ((addr & 0xfffff000) == 0xfffcc000) {
			//data = uart_read(state, addr);
 			break; 
		}
		fprintf(stderr,"NumInstr %llu, io_read_word unknown addr(0x%08x) = 0x%08x\n", state->NumInstrs, addr, data);
		SKYEYE_OUTREGS(stderr);
		//ARMul_Debug(state, 0, 0);
		break;
	}
	return data; 													
}

void at91_io_write_byte(ARMul_State *state, ARMword addr, ARMword data)
{

          unsigned char offset=0;
          unsigned char ret;
          if(addr>=io.net_io_baseaddr&&addr<(io.net_io_baseaddr+256)){
                offset=addr-io.net_io_baseaddr;
                if (data>256) {
                   fprintf(stderr,"skyeye: write data to nic is bigger than 256\n");
		   SKYEYE_OUTREGS(stderr);
                }
                ret=(unsigned char)data;
	        //chy 2003-04-26, only support one card, so the nic index must be 0
                nic_write(0, state,offset>>2,ret);
                return;
           }else{
			printf("SKYEYE: at91_io_write_byte error\n");
			SKYEYE_OUTREGS(stderr);
			exit(-1);
           }
}

void at91_io_write_halfword(ARMul_State *state, ARMword addr, ARMword data)
{
	printf("SKYEYE: at91_io_write_halfword error\n");
	SKYEYE_OUTREGS(stderr);
	exit(-1);
}

void at91_io_write_word(ARMul_State *state, ARMword addr, ARMword data)
{
	/*
  	 * The Atmel system registers
  	 */
	ARMword 	tmp;
	if(addr >= io.net_io_baseaddr && addr < (io.net_io_baseaddr + io.net_io_size)){
		printf("SKYEYE: at91_io_write_word error, shoud use write_byte for net\n");
		SKYEYE_OUTREGS(stderr);
		exit(-1);
	}
#if 0
	{
          unsigned char offset=0;
          unsigned char ret;
          if(addr>=io.net_io_baseaddr&&addr<(io.net_io_baseaddr+256)){
                offset=addr-io.net_io_baseaddr;
                if (data>256) {
                   fprintf(stderr,"skyeye: write data to nic is bigger than 256\n");
                }
                ret=(unsigned char)data;
	        //chy 2003-04-26, only support one card, so the nic index must be 0
                nic_write(0, state,offset>>2,ret);
                return;
          }
        }
#endif
	switch (addr) {
	case 0xfffff108: /* ISR */
		DBG_PRINT("write ISR=0x%x\n", data);
		io.intsr = fix_int(data);
		break;
	case 0xfffff110: /* IMR */
		io.intmr = fix_int(data);
		DBG_PRINT("write IMR=%x,%x\n", data,io.intsr);
		break;
	case 0xfffff114: /* CORE interrupt status register */
		DBG_PRINT("write CISR=0x%x\n", data);
		/* io.intsr = fix_int(data);*/
		break;
	case 0xfffff120: /* IECR */
		DBG_PRINT("IECR=0x%x\n", data);
		io.intmr |= fix_int(data);
		break;
	case 0xfffff124: /* IDCR */
		DBG_PRINT("IDCR=0x%x\n", data);
		io.intmr &= ~fix_int(data);
		break;
	case 0xfffff128: /* CLEAR interrupts */
		DBG_PRINT("ICCR=0x%x\n", data);
		io.intsr &= ~fix_int(data);
		break;
	case 0xfffff130: /* EOI */
		DBG_PRINT(stderr, "EOI=0x%x\n", data);
		io.intsr &= ~fix_int(data);
		at91_update_int(state);
		break;
	case 0xfff00000: /* CPU ID */
		break;
	case 0xfffe00c0: /* TIMER 1 BCR */
		DBG_PRINT("T1-BCR io_write_word(0x%08x) = 0x%08x\n", addr, data);
		break;
	case 0xfffe00c4: /* TIMER 1 BMR */
		DBG_PRINT("T1-BMR io_write_word(0x%08x) = 0x%08x\n", addr, data);
		break;
	case 0xfffe0040: /* TIMER 1 CCR */
		DBG_PRINT("T1-CCR io_write_word(0x%08x) = 0x%08x\n", addr, data);
		if (data & 0x2) {
			io.syscon &= ~TC1M;
		} else {
			io.syscon |= TC1M;
		}
		break;
	case 0xfffe0044: /* TIMER 1 CMR */
		DBG_PRINT("T1-CMR io_write_word(0x%08x) = 0x%08x\n", addr, data);
		break;
	case 0xfffe0068: /* TIMER 1 IDR */
		DBG_PRINT("T1-IDR io_write_word(0x%08x) = 0x%08x\n", addr, data);
		break;
	case 0xfffe0050: /* TIMER 1 CV */
		DBG_PRINT("T1-CV io_write_word(0x%08x) = 0x%08x\n", addr, data);
		io.tcd[0] = io.tcd_reload[0] = data & 0xffff;
		break;
	case 0xfffe005c: /* TIMER 1 RC */
		DBG_PRINT("T1-RC io_write_word(0x%08x) = 0x%08x\n", addr, data);
		io.tcd[0] = io.tcd_reload[0] = data & 0xffff;
		break;
	case 0xfffe0060: /* TIMER 1 SR */
		DBG_PRINT("T1-SR io_write_word(0x%08x) = 0x%08x\n", addr, data);
		break;
	case 0xfffe0064: /* TIMER 1 IER */
		DBG_PRINT("T1-IER io_write_word(0x%08x) = 0x%08x\n", addr, data);
		break;
	/* zy 2004-4-13 */
	case 0xffc00100: /*SYSCON       */ /* you may like to change it ,but it's only for simulation*/
		printf("\nat91_io_write_word():SYSCON=0x%x",data);
		tmp = io.syscon;
		io.syscon = data;
		/*
	        if ((tmp & LCDEN) != (data & LCDEN)) {
			printf("\nwill call function at91_update_lcd()");
		        at91_update_lcd(state);
		}
		*/
		break;
	case 0xffc002c0: /*LCDCON       */ /* because AT91 hasn't LCD support ,so it's just for simulation*/
	                                 /* you can change this address to what may be suitable for your board*/
		printf("\nat91_io_write_word():LCDCON=0x%x",data);
		tmp = io.lcdcon;
                io.lcdcon = data;
	        if ((tmp & (VBUFSIZ|LINELEN|GSEN|GSMD)) != (data & (VBUFSIZ|LINELEN|GSEN|GSMD))) {
			printf("\nwill call function at91_update_lcd()");
		        at91_update_lcd(state);
		}
		break;
	//	
	case 0xffc00240: //INTSR
		io.intsr=data;
		break;								
	default:
		if ((addr & 0xfffff000) == 0xfffd0000) {
			uart_write(state, addr, data);
			break;
		}
		if ((addr & 0xfffff000) == 0xfffcc000) {
			/* uart_write(state, addr, data);*/
			break;
		}
		if ((addr & 0xffffff00) == 0xfffff000)
			break;
		if ((addr & 0xffff0000) == 0xffff0000) {
			DBG_PRINT("io_write_word(0x%08x) = 0x%08x\n", addr, data);
			break;
		}
		//DBG_PRINT("io_write_word(0x%08x) = 0x%08x\n", addr, data);
		fprintf(stderr,"NumInstr %llu,io_write_word unknown addr(0x%08x) = 0x%08x\n", state->NumInstrs, addr, data);
		SKYEYE_OUTREGS(stderr);
		//ARMul_Debug(state, 0, 0);
		break;
	}
}

void at91_mach_init(ARMul_State *state, machine_config_t *this_mach)
{
	//chy 2003-08-19, setprocessor
	ARMul_SelectProcessor(state, ARM_v4_Prop);

	this_mach->mach_io_do_cycle = 		at91_io_do_cycle;
	this_mach->mach_io_reset = 		at91_io_reset;
	this_mach->mach_io_read_byte = 		at91_io_read_byte;
	this_mach->mach_io_write_byte = 	at91_io_write_byte;
	this_mach->mach_io_read_halfword = 	at91_io_read_halfword;
	this_mach->mach_io_write_halfword = 	at91_io_write_halfword;
	this_mach->mach_io_read_word = 		at91_io_read_word;
	this_mach->mach_io_write_word = 	at91_io_write_word;

	this_mach->mach_update_int = 		at91_update_int;

	//ksh 2004-2-7


	state->mach_io.instr = (ARMword *)&io.intsr;
	//*state->io.instr = (ARMword *)&io.intsr;
	//state->io->net_flags = (ARMword *)&io.net_flags;
	state->mach_io.net_int = (ARMword *)&io.net_int;
 	/* zy 2004-4-13 */
	this_mach->mach_lcd_write       =at91_lcd_write;
	state->mach_io.lcd_is_enable    =(ARMword *)&io.lcd_is_enable;
	state->mach_io.lcd_addr_begin   =(ARMword *)&io.lcd_addr_begin;
	state->mach_io.lcd_addr_end     =(ARMword *)&io.lcd_addr_end;
	//
	/*ywc 2004-04-01*/
	state->mach_io.ts_int		=(ARMword *)&io.ts_int;
	state->mach_io.ts_is_enable	=(ARMword *)&io.ts_is_enable;	
	state->mach_io.ts_addr_begin	=(ARMword *)&io.ts_addr_begin;	
	state->mach_io.ts_addr_end	=(ARMword *)&io.ts_addr_end;
	state->mach_io.ts_buffer	=(ARMword *)&io.ts_buffer[0];
}
		
