#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/select.h>

#include "armdefs.h"
#include "pxa.h"

/*ywc 2005-03-30*/
#include "skyeye_flash.h"

#define F_CORE (100 * 1024 * 1024) //core frequence
#define F_RTC 32768					//RTC
#define F_OS	3686400				//OS timer
#define RT_SCALE (F_CORE / F_RTC)
#define OS_SCALE (F_CORE / F_OS / 10)

#define FF_SCALE	200		//FF UART

//ywc,2004-11-30,declare a external array which stores touchscreen event information
extern unsigned long Pen_buffer[8]; // defined in skyeye_lcd.c
//ywc,2004-11-30,declare a external array which stores touchscreen event information,end

//chy:  lubbock, cerf, idp are different board
// I can add a union to cover the difference in the future
//chy:  refer to linux/include/asm-arm/arch-pxa/pxa-regs.h
typedef struct pxa_io_t{
	/*interrupt controller*/
	u32 icpr;
	u32 icip;
	u32 icfp;
	u32 icmr;
	u32 iccr;
	u32 iclr;

	/*real time clock(RTC)*/
	u32 rcnr;
	u32 rtar;
	u32 rtsr;
	u32 rttr;
	//chy: maybe not regs ????
	u32 rt_scale; /*core frequence to 32.768K*/
	u32 rt_count; 

	/*os timer*/
	u32 oscr;
	u32 osmr0, osmr1, osmr2, osmr3;
	u32 ower;
	u32 ossr;
	u32 oier;
	//chy: maybe not regs ????
	u32 os_scale; //chy: I can not find it in pxa-regs.h????
	
	/*full function uart controller*/
	u32 ffrbr;
	u32 ffthr;
	u32 ffier;
	u32 ffiir;
	u32 fffcr;
	u32 fflcr;
	u32 ffmcr;
	u32 fflsr;
	u32 ffmsr;
	u32 ffspr;
	u32 ffisr;
	u32 ffdll;
	u32 ffdlh;

	u32 ff_scale;
	/*bluetooth function uart controller*/
	u32 btrbr;
	u32 btthr;
	u32 btier;
	u32 btiir;
	u32 btfcr;
	u32 btlcr;
	u32 btmcr;
	u32 btlsr;
	u32 btmsr;
	u32 btspr;
	u32 btisr;
	u32 btdll;
	u32 btdlh;
	/*standard uart controller*/
	u32 strbr;
	u32 stthr;
	u32 stier;
	u32 stiir;
	u32 stfcr;
	u32 stlcr;
	u32 stmcr;
	u32 stlsr;
	u32 stmsr;
	u32 stspr;
	u32 stisr;
	u32 stdll;
	u32 stdlh;
	/*core clock*/
	u32 cccr;
	u32 cken;
	u32 oscc;
	
	//ywc,2004-11-30,add io of LCD and Touchscreen
	/*LCD*/
	u32 lccr0;
	u32 lccr1;
	u32 lccr2;
	u32 lccr3;
	
	u32 fdadr0;
	u32 fdadr1;
	
	u32 fsadr0;
	u32 fsadr1;

	u32 lcd_is_enable;
	u32 lcd_addr_begin;
	u32 lcd_addr_end;
	
	/*TouchScreen*/
	u32	ts_int;                     
	u32	ts_buffer[8];
	u32 	ts_is_enable;
	u32 	ts_addr_begin;
	u32 	ts_addr_end;
	//ywc,2004-11-30,add io of LCD and Touchscreen,end
	
}pxa_io_t;

static pxa_io_t pxa_io;

//ywc,2004-11-30,add 4 functions for pxa's LCD simulation
static void pxa_lcd_write(ARMul_State *state, ARMword addr, ARMword data)
{
        lcd_write(state, addr, data);
}

static void pxa_lcd_disable(ARMul_State *state)
{
        lcd_disable(state);
}

static void pxa_lcd_enable(ARMul_State *state, int width, int height, int depth)
{
   //printf("\n pxa_update_lcd: will call lcd_enable !");
   //printf("\n");
   lcd_enable(state, width, height, depth);
   //printf("\n pxa_update_lcd: end call lcd_enable !");
   //printf("\n");
}

static void pxa_update_lcd(ARMul_State *state)
{
	int width, height, depth;
       // pxa_lcd_disable(state);
		
	width  = (pxa_io.lccr1 & LCCR1_PPL) + 1;	
	height = ((pxa_io.lccr2 & LCCR2_LPP) + 1)*2;	
	depth  = 1 << ((pxa_io.lccr3 & LCCR3_BPP)>>24);
		
		/*
                depth = 16;
               	width = 640;
                height = 480;
		*/
        pxa_lcd_enable(state, width, height, depth);
}
//ywc,2004-11-30,add 4 functions for pxa's LCD simulation,end

static void refresh_irq(ARMul_State *);
static void pxa_io_reset()
{
	memset(&pxa_io, 0, sizeof(pxa_io));
	//chy 2003-08-25
	pxa_io.cccr = 0x121; // 1 0010 0001
	pxa_io.cken = 0x17def;
	
	//ywc,2004-11-30,evaluate io of LCD and Touchscreen
	pxa_io.lcd_addr_begin  	=0xd0000000;
	pxa_io.lcd_addr_end   	=0xd0000000;
	
	pxa_io.ts_int		=1<<15; 
	pxa_io.ts_is_enable	=1;
	pxa_io.ts_addr_begin  	=0x40000300;
	pxa_io.ts_addr_end    	=0x4000031f; 
	//ywc,2004-11-30,evaluate io of LCD and Touchscreen,end
	
};
void pxa_io_write_byte(ARMul_State *state, ARMword addr, ARMword data)
{
        printf("SKYEYE: pxa_io_write_byte error\n");
        exit(-1);
}

void pxa_io_write_halfword(ARMul_State *state, ARMword addr, ARMword data)
{
        printf("SKYEYE: pxa_io_write_halfword error\n");
        exit(-1);
}

//ywc,2004-11-30,add for pxa's LCD simulation
static int once=0;
static u32 * fdadr0ADDRESS;
//ywc,2004-11-30,add for pxa's LCD simulation,end
static void
pxa_io_write_word(ARMul_State *state, ARMword addr, ARMword data)
{
	pxa_ioregnum_t  ioregaddr=addr;
	
	//ywc,2004-11-30,add for pxa's LCD simulation
	u32 tmp;
        mem_bank_t *mbp;
	//ywc,2004-11-30,add for pxa's LCD simulation,end

	//printf("SKYEYE:pxa_io_write_word: io addr 0x%x, RCNR 0x%x\n",ioregaddr,RCNR);

	switch(ioregaddr){
		/*RTC*/
		case RCNR:
			pxa_io.rcnr = data;
			break;	
		case RTAR:
			pxa_io.rtar = data;
			break;
		case RTSR:
			pxa_io.rtsr |= (data & 0xc);
			pxa_io.rtsr &= ~(data & 0x3);
			break;
		case RTTR:
			pxa_io.rttr = data & 0x3ffffff;
			break;
		/*OS timer*/
		case OSCR:
			pxa_io.oscr = data;
			break;
		case OSMR0:
			pxa_io.osmr0 = data;
			break;
		case OSMR1:
			pxa_io.osmr1 = data;
			break;
		case OSMR2:
			pxa_io.osmr2 = data;
			break;
		case OSMR3:
			pxa_io.osmr3 = data;
			break;
		case OWER:
			pxa_io.ower |= data & 0x1;
			break;
		case OSSR:
			pxa_io.ossr &= ~(data & 0xf);
			break;
		case OIER:
			pxa_io.oier = data & 0xf;
			break;

		/*interrupt control*/
		//ywc,2004-11-30,for touchscreen use ICPR
		case ICPR:
			pxa_io.icpr = data;
			//printf("\n write ICPR=%x",pxa_io.icpr);
			//printf("\n");
			break;
		//ywc,2004-11-30,for touchscreen use ICPR,end
			
		//case ICIP:
		//case ICFP:
			/*read only*/
		//	break;
		case ICMR:
			pxa_io.icmr = data;
			break;
		case ICLR:
			pxa_io.iclr = data;
			break;

		//ffuart contril
                case FFTHR:    
			{       /*static tx_cnt=0;*/
                                unsigned char c = data;
                                write(skyeye_config.uart.fd_out, &c, 1);
				/*chy 2004-07-21 from luzhetao: after write char, should set some value*/
			 	pxa_io.ffiir&=~0x2;
                                pxa_io.ffiir|=0x1;
                                pxa_io.fflsr&=~0x60;
				//tx_cnt++;
				//pxa_io.ffier|=0x2;
				/*
				if(tx_cnt>63){
				  tx_cnt=0;
				  pxa_io.ffiir|=0x2;
				  pxa_io.fflsr|=0x60;
				  refresh_irq(state);
				}
				*/
				//printf("SKYEYE: write FFTHR %x, iir %x,lsr %x, ier %x\n",data, pxa_io.ffiir,pxa_io.fflsr,pxa_io.ffier);
                        }
                        break;
		case FFIER:
			pxa_io.ffier = data & 0xff; 
			break;
		case FFFCR:  //write only
			pxa_io.fffcr = data & 0xc7 ;
			//pxa_io.ffiir = data & 0xc7 ;
			//if(pxa_io.fffcr & 2
			//printf("SKYEYE: write FFFCR %x, but is ffiir %x\n",pxa_io.fffcr,pxa_io.ffiir);
			break;
		case FFLCR:
			pxa_io.fflcr = data & 0xff;
			break;
		//core clock 
		case CCCR:
			pxa_io.cccr = data & 0x3ff;
			break; 
		case CKEN:
			pxa_io.cken = data & 0x17def; 
			break;
		case OSCC:
			pxa_io.oscc = data & 0x3;
			break;
			
		//ywc,2004-11-30,add for pxa's LCD simulation
		case LCCR0:
			tmp=pxa_io.lccr0;
			pxa_io.lccr0 = data;
			if((!(tmp & LCCR0_ENB)) && (pxa_io.lccr0 & LCCR0_ENB)){
				if(once)
					break;
				once++;
				/*
				printf("\nFDADR0=%x,FDADR1=%x",pxa_io.fdadr0,pxa_io.fdadr1);
				printf("\n");
				pxa_io.lcd_addr_begin = pxa_io.fdadr0 + 16*(2+2);
				printf("\nlcd_addr_begin=%x",pxa_io.lcd_addr_begin);
				printf("\n");
				*/
				pxa_update_lcd(state);
			}
			break;
		case LCCR1:
			pxa_io.lccr1 = data;
			break;
		case LCCR2:
			pxa_io.lccr2 = data;
			break;
		case LCCR3:
			pxa_io.lccr3 = data;
			break;
		case FDADR0:
			pxa_io.fdadr0 = data;
			
			//printf("\nFDADR0=%x",pxa_io.fdadr0);
			//printf("\n");
			/*	
			mbp = bank_ptr(pxa_io.fdadr0); 
			if(!mbp){
			        fprintf(stderr, "No bank at address 0x%x", pxa_io.fdadr0);
                                return;
                         }   
                         fdadr0ADDRESS=(u32 *) &state->mem.rom[mbp - skyeye_config.mem.mem_banks][(pxa_io.fdadr0 - mbp->addr)];
			 printf("\n %p",fdadr0ADDRESS);
			*/
		//	 printf("\n 0: %x",*((u32 *)fdadr0ADDRESS));
		//	 printf("\n 1: %x",*((u32 *)fdadr0ADDRESS+1));
		//	 printf("\n 2: %x",*((u32 *)fdadr0ADDRESS+2));
		//	 printf("\n 3: %x",*((u32 *)fdadr0ADDRESS+3));
			 //printf("\n");
			 	
			 break;
		case FDADR1:
			pxa_io.fdadr1 = data;
			//printf("\nFDADR1=%x",pxa_io.fdadr1);
			//printf("\n");
			break;
		case FSADR0:
                        pxa_io.fsadr0 = data;
			//printf("\nFSADR0=%x",pxa_io.fsadr0);
			//printf("\n");
                        break;
		case FSADR1:
                        pxa_io.fsadr1 = data;
			//printf("\nFSADR1=%x",pxa_io.fsadr1);
			//printf("\n");
                        break;
		//ywc,2004-11-30,add for pxa's LCD simulation,end
#if 0		
		case FFIIR:  //read only
		case FFMCR:  //no use			
		case FFMSR:  //read only no use
		case FFSPR:  // no use
		case FFISR:  // no use
#endif
	
		default:
			//chy 2003-09-03 if debug, uncommit it
			//log_msg("SKYEYE: pxa_io_write_word: unknown addr 0x%x, reg15 0x%x \n", addr,state->Reg[15]);
			;
	};

};

ARMword pxa_io_read_byte(ARMul_State *state, ARMword addr)
{
        printf("SKYEYE: pxa_io_read_byte error\n");
        exit(-1);
}
ARMword pxa_io_read_halfword(ARMul_State *state, ARMword addr)
{
        printf("SKYEYE: pxa_io_read_halfword error\n");
        exit(-1);
}

ARMword pxa_io_read_word(ARMul_State *state, ARMword addr)
{
	u32 data;
	pxa_ioregnum_t  ioregaddr=addr;
	
	//ywc,2004-11-30,add for pxa's Touchscreen simulation
	u32 ts_addr;
	//ywc,2004-11-30,add for pxa's Touchscreen simulation,end

	//printf("SKYEYE:pxa_io_read_word: io addr 0x%x\n",ioregaddr);
	
	//ywc 2004-11-30 read the touch srceen data buffer,return to the ts device driver
	ts_addr=addr & ~3;   // 1 word==4 byte
	if(ts_addr >= pxa_io.ts_addr_begin && ts_addr <= pxa_io.ts_addr_end){
		data=pxa_io.ts_buffer[(ts_addr-pxa_io.ts_addr_begin)/4];
	        //        printf("addr=%p,ts_addr=%p\n",addr,ts_addr);
                //printf("&io.ts_buffer[%d]=%p\n",(ts_addr-io.ts_addr_begin)/4,&io.ts_buffer[(ts_addr-io.ts_addr_begin)/4]);
          //      printf("data=%d\n",data);
		return data;
	}
	//ywc 2004-11-30 read the touch srceen data buffer,return to the ts device driver,end
	
	switch(addr){
		/*RTC*/
		case RCNR:
			data = pxa_io.rcnr;
			break;	
		case RTAR:
			data = pxa_io.rtar;
			break;
		case RTSR:
			data = pxa_io.rtsr;
			break;
		case RTTR:
			data = pxa_io.rttr;
			break;
			
	/*OS timer*/
		case OSCR:
			data = pxa_io.oscr;
			break;
		case OSMR0:
			data = pxa_io.osmr0;
			break;
		case OSMR1:
			data = pxa_io.osmr1;
			break;
		case OSMR2:
			data = pxa_io.osmr2;
			break;
		case OSMR3:
			data = pxa_io.osmr3;
			break;
		case OWER:
			data = pxa_io.ower;
			break;
		case OSSR:
			data = pxa_io.ossr;
			break;
		case OIER:
			data = pxa_io.oier;
			break;
		
		/*interrupt controler*/	
		case ICPR:
			data = pxa_io.icpr;
			break;
		case ICIP:
			data  = (pxa_io.icmr & pxa_io.icpr) & ~pxa_io.iclr; 
			break;
		case ICFP:
			data = (pxa_io.icmr & pxa_io.icpr) & pxa_io.iclr;
			break;
		case ICMR:
			data = pxa_io.icmr;
			break;
		case ICLR:
			data = pxa_io.iclr;
			break;

		/* ffuart control */
		case FFRBR:   //chy: 2003-08-24 have some question, maybe skyeye will be locked here ????
			{
                                //unsigned char c;
                                //read(skyeye_config.uart.fd_in, &c, 1);
                                //data = c;
							data = pxa_io.ffrbr&0xff;
							pxa_io.icpr &= ~FFUART_IRQ;
							pxa_io.icip &= ~FFUART_IRQ;
							pxa_io.ffiir &= ~0x4;
							/*chy 2004-07-21 from lzt, afte read char, should set this bit*/
							pxa_io.ffiir |= 0x1;

			//printf("SKYEYE: read FFRBR, but set ffiir  and ~04, now %x\n",pxa_io.ffiir);
							pxa_io.fflsr &= ~0x1;
																													
                        };
			break;
		case FFIER:
			data=pxa_io.ffier; 
			break;
		case FFIIR:  //read only
			data=pxa_io.ffiir & 0xcf ;
			pxa_io.icpr &= ~FFUART_IRQ;
                        pxa_io.icip &= ~FFUART_IRQ;
			//printf("SKYEYE: read FFIIR  %x\n",data);
			break;
		case FFLCR:
			data=pxa_io.fflcr;
			break;
		case FFLSR:  //read only 
			//chy 2003-08-24  NOTICE!
			//fixed by ksh 2004-05-09,for FFUART can input char,I am not sure?
			//pxa_io.fflsr=1<<5; // LSR's TDRQ =1 , then the cpu can write char again
			pxa_io.fflsr|=0x60; // chy 2003-09-02 for LSR_TEMT|LSR_THRE 0x40|0x20
			data=pxa_io.fflsr & 0xff ;		
			break;

		// core clock 
		case CCCR: 
			data=pxa_io.cccr;
			break;	
		case CKEN:
			data=pxa_io.cken;
			break;
		case OSCC:
			data=pxa_io.oscc;
			break;
		//ywc,2004-11-30,add for pxa's LCD simulation
		case LCCR0:
			data=pxa_io.lccr0;
			break;
		case LCCR1:
			data=pxa_io.lccr1;
			break;
		case LCCR2:
			data=pxa_io.lccr2;
			break;
		case LCCR3:
			data=pxa_io.lccr3;
			break;
		case FDADR0:
                        data=pxa_io.fdadr0;
                        break;
                case FDADR1:
                        data=pxa_io.fdadr1;
                        break;
		case FSADR0:
                        data=pxa_io.fsadr0;
                        break;
		case FSADR1:
                        data=pxa_io.fsadr1;
                        break;
		//ywc,2004-11-30,add for pxa's LCD simulation,end
#if 0
		case FFFCR:  //write only
		case FFMCR:  //no use			
		case FFMSR:  //read only no use
		case FFSPR:  // no use
		case FFISR:  // no use
#endif

		default:
			data = 0;
			//chy 2003-09-03 if debug, uncommit it
			//log_msg("SKYEYE: pxa_io_read_word: unknown addr 0x%x, reg15 0x%x \n", addr,state->Reg[15]);
	};
	
	return data;
};

//ywc,2004-11,30,add for lcd cycle
static int lcd_cycle_scale=0;
static int LCD_CYCLE_SCALE=400;
//ywc,2004-11,30,add for lcd cycle,end
static void
pxa_io_do_cycle(ARMul_State *state)
{
	/*RTC*/
	if (++pxa_io.rt_scale >= RT_SCALE){
		pxa_io.rt_scale = 0;
		if (pxa_io.rt_count ++ == (pxa_io.rttr & 0xffff)){
			pxa_io.rt_count = 0;
			
			if (pxa_io.rcnr ++ == pxa_io.rtar){
				if (pxa_io.rtsr & 0x4){
					pxa_io.rtsr |= 0x1;
				};
			}
			if (pxa_io.rtsr & 0x8){
				pxa_io.rtsr |= 0x2;
			}
		}
	};

	/*OS timer*/
	if (++pxa_io.os_scale >= OS_SCALE){
		u32 mask = 0;
		u32 count;
		
		pxa_io.os_scale = 0;
		count = pxa_io.oscr++;
		
		if (count == pxa_io.osmr0)
			mask = 1;
		if (count == pxa_io.osmr1)
			mask |= 0x2;
		if (count == pxa_io.osmr2)
			mask |= 0x4;
		if (count == pxa_io.osmr3){
			mask |= 0x8;
			if (pxa_io.ower & 1){
				state->NresetSig = LOW;
			   	printf("************SKYEYE: WatchDog reset!!!!!!!**************\n");	
			}
		}
		pxa_io.ossr |= mask;
	}

	/*FF UART */
	//pxa_io.utsr0 = 1; /*always TFS, no others*/
	//pxa_io.utsr1 = 0x4; /*TNF*/
	if (++pxa_io.ff_scale >= FF_SCALE){
		pxa_io.ff_scale = 0;
		if(!(FFUART_IRQ&pxa_io.icpr))
		{
			fd_set rset;
			struct timeval tv;
		
			FD_ZERO(&rset);
			FD_SET(0, &rset);
			tv.tv_sec = 0;
			tv.tv_usec = 0;
			if (select(1, &rset, NULL, NULL, &tv) == 1){
				//pxa_io.utsr1 |= 0x2;	//RNE
				//pxa_io.utsr0 |= 0x4; //Rx idle
				//added by ksh,2004-4-25,get uart data and  set the interrupt
				unsigned char c;
				read(skyeye_config.uart.fd_in, &c, 1);
				pxa_io.ffrbr = (int)c;
				pxa_io.ffiir |= 0x4; //Rx idle
			        //printf("SKYEYE: io_do_cycle  set ffiir  or 04, now %x\n",pxa_io.ffiir);
				pxa_io.fflsr |= 0x01; //Data ready
			                                                            
			}
		}
		//ywc,2004-11-30,check and get the mouse event
		if(!(pxa_io.icpr&pxa_io.ts_int)){ //if now has no ts interrupt,then query 
	              if(Pen_buffer[6]==1){  //should trigger a interrupt
				*(state->mach_io.ts_buffer+0)=Pen_buffer[0];
				*(state->mach_io.ts_buffer+1)=Pen_buffer[1];
		//		*(state->mach_io.ts_buffer+2)=Pen_buffer[2];
		//		*(state->mach_io.ts_buffer+3)=Pen_buffer[3];
				*(state->mach_io.ts_buffer+4)=Pen_buffer[4];
		//		*(state->mach_io.ts_buffer+5)=Pen_buffer[5];
				*(state->mach_io.ts_buffer+6)=Pen_buffer[6];
		//		*(state->mach_io.ts_buffer+7)=Pen_buffer[7];
//			    printf("\nbefore set ENIT2 bit,pxa_io.icpr=%x",pxa_io.icpr);	
//			    printf("\n");	
			
			pxa_io.icpr|= pxa_io.ts_int;//set EINT2 bit to trigger a interrupt,ts driver will clear it
//			    printf("\nafter set ENIT2 bit,pxa_io.icpr=%x",pxa_io.icpr);	
//			    printf("\n");	
       				
				Pen_buffer[6]=0;
        		}
		 }
		 //ywc,2004-11-30,check and get the mouse event,end
	}
	
	//ywc,2004-11-30,for lcd cycle
	if (++lcd_cycle_scale >=LCD_CYCLE_SCALE){
		lcd_cycle_scale=0;
		lcd_cycle(state);	
	}
	//ywc,2004-11-30,for lcd cycle,end
	
	/*reset interrupt pin status*/
	refresh_irq(state);
};

static void refresh_irq(ARMul_State *state)
{
	u32 irq = 0;
	u32 mask;
	
	/*RTC*/
	if ((pxa_io.rtsr & 0x1) && (pxa_io.rtsr & 0x4))
		irq |= RTC_ALARM_IRQ;
	if ((pxa_io.rtsr & 0x2) && (pxa_io.rtsr & 0x8))
		irq |= RTC_HZ_IRQ;
	/*OS time*/
	mask = pxa_io.oier & pxa_io.ossr;
	irq |= (mask << OS_IRQ_SHF);
#if 0	
	/*UART3*/
	if ((pxa_io.utcr3 & 0x8) && (pxa_io.utsr0 & 0x6))
	   irq |= UART3_IRQ;
	if (pxa_io.utsr0 & 0x38)
		irq |= UART3_IRQ;
	
	if ((pxa_io.utcr3 & 0x10) && (pxa_io.utsr0 & 0x1))
		irq |= UART3_IRQ;
#endif	
	/*FFUART*/
	//chy 2004-07-19 luzetao find a solution: if (pxa_io.ffier&0x2){ irq |= FFUART_IRQ; }
	//chy 2004-07-19 then linux-2.6.x can run correctly!
	//chy 2004-06-23  FFIER    TIE or RAVIE, so shoud be 3, FFIIR 0x4 or 0x2=0x6   
	
	/* chy 2004-07-21 from luzhetao
	if((pxa_io.ffier&0x02)||((pxa_io.ffier&0x3) && (pxa_io.ffiir&0x6))){
						    irq |= FFUART_IRQ;
		}
	*/	
	/*chy 2004-07-21 from luzhetao: produce a IRQ*/
        if((pxa_io.ffier&0x1) && (pxa_io.ffiir&0x4)){
                        irq |= FFUART_IRQ;
                        pxa_io.ffiir &=~0x1;
        }
	/*chy 2004-07-21 from luzhetao: produce a IRQ and ox2 should clean by OS driver */
        if (pxa_io.ffier&0x2){
                irq |= FFUART_IRQ;
                pxa_io.ffiir|=0x2;
                pxa_io.ffiir&=~0x1;
                pxa_io.fflsr|=0x60;
        }

	//ywc,2004-11-30,produce a touchsreen IRQ
	if (pxa_io.icpr&pxa_io.ts_int){
		irq |= pxa_io.ts_int;	
	}
	//ywc,2004-11-30,produce a touchsreen IRQ,end
	
	pxa_io.icpr = irq;
	pxa_io.icip = (pxa_io.icmr & pxa_io.icpr) & ~pxa_io.iclr; 
	pxa_io.icfp = (pxa_io.icmr & pxa_io.icpr) & pxa_io.iclr;
	state->NirqSig = pxa_io.icip ? LOW:HIGH; 
	state->NfiqSig = pxa_io.icfp ? LOW:HIGH;
}

void pxa_mach_init(ARMul_State *state, machine_config_t *mc)
{
        //chy 2003-08-19, setprocessor
        ARMul_SelectProcessor(state, ARM_XScale_Prop|ARM_v5_Prop|ARM_v5e_Prop);
        //chy 2004-05-09, set lateabtSig
        state->lateabtSig = LOW;

	state->Reg[1] = 89; /*lubbock machine id.*/
	pxa_io_reset();
	/*added by ksh for energy estimation in 2004-11-26 */
	state->energy.cccr = pxa_io.cccr;

	mc->mach_io_do_cycle = 		pxa_io_do_cycle;
	mc->mach_io_reset = 		pxa_io_reset;
	mc->mach_io_read_byte =         pxa_io_read_byte;
        mc->mach_io_write_byte =        pxa_io_write_byte;
        mc->mach_io_read_halfword =     pxa_io_read_halfword;
        mc->mach_io_write_halfword =    pxa_io_write_halfword;
	mc->mach_io_read_word = 	pxa_io_read_word;
	mc->mach_io_write_word = 	pxa_io_write_word;
	
	//ywc,2004-11-30, for pxa's LCD and Touchscreen simulation
	mc->mach_lcd_write		=pxa_lcd_write;
	state->mach_io.lcd_is_enable	=(ARMword *)&pxa_io.lcd_is_enable;	
	state->mach_io.lcd_addr_begin	=(ARMword *)&pxa_io.lcd_addr_begin;	
	state->mach_io.lcd_addr_end	=(ARMword *)&pxa_io.lcd_addr_end;
	
	state->mach_io.ts_int		=(ARMword *)&pxa_io.	ts_int;
	state->mach_io.ts_is_enable	=(ARMword *)&pxa_io.ts_is_enable;	
	state->mach_io.ts_addr_begin	=(ARMword *)&pxa_io.ts_addr_begin;	
	state->mach_io.ts_addr_end	=(ARMword *)&pxa_io.ts_addr_end;
	state->mach_io.ts_buffer	=(ARMword *)&pxa_io.ts_buffer[0];
	//ywc,2004-11-30, for pxa's LCD and Touchscreen simulation
	
	/*ywc 2005-03-30 for flash simulation*/
	mc->mach_flash_read_byte =      Intel28F128J3A_flash_read_byte;
	mc->mach_flash_write_byte =     Intel28F128J3A_flash_write_byte;
	mc->mach_flash_read_halfword =  Intel28F128J3A_flash_read_halfword;
	mc->mach_flash_write_halfword = Intel28F128J3A_flash_write_halfword;
	mc->mach_flash_read_word =      Intel28F128J3A_flash_read_word;
	mc->mach_flash_write_word =     Intel28F128J3A_flash_write_word;
}
